/*-
 * Copyright (C)2012..2014 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2012..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: mercumeterq.c,v 1.36 2017/01/09 00:10:00 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT   3
#define Bpars_COMPAT  3
#define RECOBE_COMPAT 1

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <babolo/recobe.h>
#include <mife.h>
#include <mercumeter.h>

#define _ctr    memeter_ctr *ctr
#define _6r     memeter_6req
#define _fmt    const char *fmt

static u_char wk[] =
{  0xb2, 0xac, 0x97, 0xaf,  0x91, 0x92, 0x81, 0x98,  0x99, 0x9f, 0xb0, 0x9d,  0xb9, 0xba, 0xaa, 0xa9
,  0xbc, 0x87, 0xb5, 0xa1,  0xa8, 0x83, 0x84, 0x89,  0x96, 0x88, 0x8e, 0x94,  0x8f, 0x95, 0xa0, 0xa2
,  0X9A, 0XBE, 0XAE, 0x85,  0x9b, 0XBD, 0x90, 0x82,  0XB3, 0XBF, 0XB4, 0x86,  0x80, 0xb1, 0x8a, 0XB7
,  0X9C, 0xab, 0XB6, 0XA6,  0XAD, 0xa5, 0xbb, 0X9E,  0XA3, 0x93, 0XA4, 0xb8,  0x8d, 0x8c, 0x8b, 0XA7

,  0XE1, 0XE2, 0XF7, 0XE7,  0XE4, 0XE5, 0XF6, 0XFA,  0XE9, 0XEA, 0XEB, 0XEC,  0XED, 0XEE, 0XEF, 0XF0
,  0XF2, 0XF3, 0XF4, 0XF5,  0XE6, 0XE8, 0XE3, 0XFE,  0XFB, 0XFD, 0XFF, 0XF9,  0XF8, 0XFC, 0XE0, 0XF1
,  0XC1, 0XC2, 0XD7, 0XC7,  0XC4, 0XC5, 0XD6, 0XDA,  0XC9, 0XCA, 0XCB, 0XCC,  0XCD, 0XCE, 0XCF, 0XD0
,  0XD2, 0XD3, 0XD4, 0XD5,  0XC6, 0XC8, 0XC3, 0XDE,  0XDB, 0XDD, 0XDF, 0XD9,  0XD8, 0XDC, 0XC0, 0XD1
};

static char cfb[][21] =
{ "ScanEnable"
, "OptEnable"
, "CheckEnable"
, "ExeEnable"
, "PListReadOnly"
, "NotUsed"
, "CaptureEnable"
, "DayLightSavingEnable"
};

static char csm[][13] =
{ "PlcProtected"
, "TimerSet    "
, "None2       "
, "None3       "
, "None4       "
, "None5       "
, "None6       "
, "None7       "
};

static char ccn[][6] =
{ "None0"
, "None1"
, "None2"
, "None3"
, "None4"
, "None5"
, "None6"
, "None7"
};

static char cbx[][13] =
{ "EMPTY"
, "REQ_ACCEPTED"
, "TRANSACTION"
, "OK"
, "ERROR"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "UnKnown"
, "ANS_READY"
, "TRANS_TYPE"
, "None6"
, "None7"
};

static char wd[][4] = {"Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"};

static char mm[][4] =
{ "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};

static int c4ra0(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd));};
static int c4rx1(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 1));};
static int c4rx6(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 6));};
static int c4r10(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 0, MEMETER_C_PARAM_PWR_P));};
static int c4r11(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 1, MEMETER_C_PARAM_PWR_P));};
static int c4r12(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 2, MEMETER_C_PARAM_PWR_P));};
static int c4r13(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 3, MEMETER_C_PARAM_PWR_P));};
static int c4r14(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 0, MEMETER_C_PARAM_PWR_Q));};
static int c4r15(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 1, MEMETER_C_PARAM_PWR_Q));};
static int c4r16(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 2, MEMETER_C_PARAM_PWR_Q));};
static int c4r17(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 3, MEMETER_C_PARAM_PWR_Q));};
static int c4r18(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 0, MEMETER_C_PARAM_PWR_S));};
static int c4r19(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 1, MEMETER_C_PARAM_PWR_S));};
static int c4r20(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 2, MEMETER_C_PARAM_PWR_S));};
static int c4r21(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 3, MEMETER_C_PARAM_PWR_S));};
static int c4r23(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 3));};
static int c4r25(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_P));};
static int c4r26(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_Q));};
static int c4r27(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_S));};
static int c4r34(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_P));};
static int c4r35(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_Q));};
static int c4r36(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, MEMETER_C_PARAM_PWR_S));};
static int c4r52(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 0));};
static int c4r54(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 5));};
static int c4r61(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 6, 1));};
static int c4r62(_ctr, u_int32_t cmd) {return(_6r(ctr, cmd, 1, 6));};

static void par0(_fmt, u_int32_t *ans) {printf(fmt, ans[0]);}
static void par1(_fmt, u_int32_t *ans) {printf(fmt, ans[0], ans[1]);}
static void par2(_fmt, u_int32_t *ans) {printf(fmt, ans[0], ans[1], ans[2]);}
static void par3(_fmt, u_int32_t *ans) {printf(fmt, ans[0], ans[1], ans[2], ans[3]);}
static void par4(_fmt, u_int32_t *ans) {printf(fmt, ans[0], ans[1], ans[2], ans[3], ans[4]);}
static void par5(_fmt, u_int32_t *ans) {printf(fmt, ans[0], ans[1], ans[2], ans[3], ans[4], ans[5]);}
static void par7(_fmt, u_int32_t *ans) {
    printf(fmt, ans[0], ans[1], ans[2], ans[3], ans[4], ans[5], ans[6], ans[7]);
}
static void pac0(_fmt, u_int32_t *ans) {
    u_int32_t k;

    for (k = 0; k < ans[0]; ++k) {
        printf( "%s %10u %10u %10u %10u"
              , k ? "\n             " : fmt
              , ans[k * 4 + 1], ans[k * 4 + 2], ans[k * 4 + 3], ans[k * 4 + 4]
              )
        ;
    }
    printf("\n");
}
static void pac8(_fmt, u_int32_t *ans) {
    u_int32_t k;

    for (k = 0; k < ans[0]; ++k) {
        printf( "%s %8u %8u %8u %8u"
              , k ? "\n                " : fmt
              , ans[k * 4 + 1], ans[k * 4 + 2], ans[k * 4 + 3], ans[k * 4 + 4]
              )
        ;
    }
    printf("\n");
}
static void p408(_fmt, u_int32_t *ans) {
    printf(fmt, ans[1], ans[0], ans[3], ans[2], ans[5], ans[4], ans[7], ans[6]);
}
static void p404(_fmt, u_int32_t *ans) {printf(fmt, ans[2], ans[1], ans[0], ans[5], ans[4], ans[3]);}
static void p400(_fmt, u_int32_t *ans) {printf(fmt, ans[0], 2000 + ans[3], ans[2], ans[1]);}
static void p406(_fmt, u_int32_t *a) {
    u_int32_t k;
    memeter_ansb *ans = (memeter_ansb *)a;

    printf(fmt);
    for (k = 0; k < ans->a1; ++k) printf(" %02X", ans->a2[k]);
    printf("\n");
}
static void p407(_fmt, u_int32_t *ans) {
    u_int32_t k;

    printf(fmt);
    for (k = 0; k < 4; ++k) printf(" %02X", ans[k]);
    printf("\n");
}
static void p443(_fmt, u_int32_t *ans) {
    printf( fmt
          , ans[0] & 0x01FFFF
          , (ans[0] & 0x80000000) ? "winterr" : "summer"
          , (ans[0] & 0x40000000) ? "yes" : "no"
          , (ans[0] & 0x20000000) ? "yes" : "no"
          , (ans[0] & 0x10000000) ? "yes" : "no"
          )
    ;
    printf( ", @ 20%02x-%02x-%02x %02x:%02x mean time %u\n"
          , ans[5], ans[4], ans[3], ans[1], ans[2], ans[6]
          )
    ;
}
static void p442(_fmt, u_int32_t *ans) {
    char *k = NULL;

    switch(ans[0] & 0xC0) {
    case 0x00: k = "0.2"; break;
    case 0x40: k = "0.5"; break;
    case 0x80: k = "1.0"; break;
    case 0xC0: k = "2.0"; break;
    }
    printf(fmt, k);
    switch(ans[0] & 0x30) {
    case 0x00: k = "0.2"; break;
    case 0x10: k = "0.5"; break;
    case 0x20: k = "1.0"; break;
    case 0x30: k = "2.0"; break;
    }
    printf(", C1R=%s%%", k);
    switch(ans[0] & 0x0C) {
    case 0x00: k = "57.7"; break;
    case 0x04: k = "220" ; break;
    case 0x08: k = "??"  ; break;
    case 0x0C: k = "???" ; break;
    }
    printf(", V=%sV", k);
    switch(ans[0] & 0x03) {
    case 0x00: k = "5"  ; break;
    case 0x01: k = "1"  ; break;
    case 0x02: k = "10" ; break;
    case 0x03: k = "???"; break;
    }
    printf(", I=%sA", k);
    printf(", directions=%u", (ans[1] & 0x80) ? 1 : 2);
    printf(", temperature=%uC\n", (ans[1] & 0x40) ? 40 : 20);
    printf(", mean pwr prof %s", (ans[1] & 0x20) ? "yes" : "no");
    printf(", %u phases", (ans[1] & 0x10) ? 1 : 3);
    switch(ans[1] & 0x0F) {
    case 0x00: k = "5000" ; break;
    case 0x01: k = "25000"; break;
    case 0x02: k = "1250" ; break;
    case 0x03: k = "500"  ; break;
    case 0x04: k = "1000" ; break;
    case 0x05: k = "250"  ; break;
    default  : k = "???"  ; break;
    }
    printf(", %s pulse/KWh, ph.sum %s\n", k, (ans[2] & 0x80) ? "abs" : "signed");
    printf(", %s tarificator", (ans[2] & 0x40) ? "internal" : "external");
    switch(ans[2] & 0x30) {
    case 0x00: k = "AR"  ; break;
    case 0x10: k = "A"   ; break;
    default  : k = "unkn"; break;
    }
    printf(", type %s, variant %u, mem3=%uKx8\n", k, ans[2] & 0x0F, (ans[3] & 0x80) ? 128 : 64);
    printf(", PLM modem %s", (ans[3] & 0x40) ? "yes" : "no");
    printf(", GSM modem %s", (ans[3] & 0x20) ? "yes" : "no");
    printf(", IRDA %s", (ans[3] & 0x10) ? "yes" : "no");
    switch(ans[3] & 0x0C) {
    case 0x00: k = "CAN"  ; break;
    case 0x04: k = "RS485"; break;
    case 0x08: k = "res"  ; break;
    case 0x0C: k = "no"   ; break;
    }
    printf(", If1=%s, ext supl %s\n", k, (ans[3] & 0x02) ? "yes" : "no");
    printf(", electric plumb %s", (ans[3] & 0x01) ? "yes" : "no");
    printf("%s", (ans[4] & 0xF0) ? ", UNKNOWN" : "");
    printf(", If2 %s", (ans[4] & 0x08) ? "yes" : "no");
    printf(", int If1 supl %s", (ans[4] & 0x04) ? "yes" : "no");
    printf(", control %s\n", (ans[4] & 0x02) ? "yes" : "no");
    printf(", separate pwr acc %s", (ans[4] & 0x01) ? "yes" : "no");
    printf("%s", ans[5] ? ", RESERV\n" : "\n");
}

static struct cmd4_t {
    int       (*req)(memeter_ctr *ctr, u_int32_t cmd);
    void      (*pr)(const char *fmt, u_int32_t *ans);
    u_int32_t   cmd;
    const char *fmt;
} cmd4[] =
{ {c4ra0, p400, MEMETER_GET_SERIAL         , "serial %08X @ %04u-%02u-%02u\n"}
, {c4ra0, par1, MEMETER_GET_TRANS_COFF     , "Ku=%u Ki=%u\n"}
, {c4ra0, par2, MEMETER_GET_VERSION        , "Version %02X.%02X.%02X\n"}
, {c4ra0, par0, MEMETER_GET_CADDR          , "Addr %02X\n"}
, {c4ra0, p404, MEMETER_GET_DST            , "DST %02x-%02x %02x %02x-%02x %02x\n"}
, {c4ra0, par0, MEMETER_GET_FLAGS          , "Flags %04X\n"}
, {c4ra0, p406, MEMETER_GET_STATUS         , "Status"}
, {c4ra0, p407, MEMETER_GET_LOCATION       , "Location"}
, {c4rx1, p408, MEMETER_GET_PEAK_TIME      , "Peak time %02u:%02u..%02u:%02u %02u:%02u..%02u:%02u\n"}
, {c4rx1, par7, MEMETER_GET_PEAK_PWR       , "Max pwr A+ %u %u A- %u %u R+ %u %u R- %u %u\n"}
, {c4r10, par1, MEMETER_GET_MAIN_1PWR      , "Pwr P 0 %1X %u\n"}
, {c4r11, par1, MEMETER_GET_MAIN_1PWR      , "Pwr P 1 %1X %u\n"}
, {c4r12, par1, MEMETER_GET_MAIN_1PWR      , "Pwr P 2 %1X %u\n"}
, {c4r13, par1, MEMETER_GET_MAIN_1PWR      , "Pwr P 3 %1X %u\n"}
, {c4r14, par1, MEMETER_GET_MAIN_1PWR      , "Pwr Q 0 %1X %u\n"}
, {c4r15, par1, MEMETER_GET_MAIN_1PWR      , "Pwr Q 1 %1X %u\n"}
, {c4r16, par1, MEMETER_GET_MAIN_1PWR      , "Pwr Q 2 %1X %u\n"}
, {c4r17, par1, MEMETER_GET_MAIN_1PWR      , "Pwr Q 3 %1X %u\n"}
, {c4r18, par1, MEMETER_GET_MAIN_1PWR      , "Pwr S 0 %1X %u\n"}
, {c4r19, par1, MEMETER_GET_MAIN_1PWR      , "Pwr S 1 %1X %u\n"}
, {c4r20, par1, MEMETER_GET_MAIN_1PWR      , "Pwr S 2 %1X %u\n"}
, {c4r21, par1, MEMETER_GET_MAIN_1PWR      , "Pwr S 3 %1X %u\n"}
, {c4rx1, par1, MEMETER_GET_MAIN_1V        , "V 1 %1X %u\n"}
, {c4r23, par1, MEMETER_GET_MAIN_1V        , "V 3 %1X %u\n"}
, {c4ra0, par1, MEMETER_GET_MAIN_1FREQ     , "F %1X %u\n"}
, {c4r25, par4, MEMETER_GET_FIXED_PWR      , "FIX Pwr P %02X %u %u %u %u\n"}
, {c4r26, par4, MEMETER_GET_FIXED_PWR      , "FIX Pwr Q %02X %u %u %u %u\n"}
, {c4r27, par4, MEMETER_GET_FIXED_PWR      , "FIX Pwr S %02X %u %u %u %u\n"}
, {c4ra0, par3, MEMETER_GET_FIXED_V        , "FIX V %02X %u %u %u\n"}
, {c4ra0, par1, MEMETER_GET_FIXED_FREQ     , "FIX F %1X %u\n"}
, {c4ra0, par4, MEMETER_GET_FIXED_COFF     , "FIX Coff %02X %u %u %u %u\n"}
, {c4ra0, par3, MEMETER_GET_FIXED_ANGLE    , "FIX Angle %02X %u %u %u\n"}
, {c4ra0, par7, MEMETER_GET_FIXED_DATE     , "FIX D %02X %02X %02X %02X %02X %02X %02X %02X\n"}
, {c4ra0, par4, MEMETER_GET_FIXED_ENERGY   , "FIX Energy %02X %u %u %u %u\n"}
, {c4r34, par4, MEMETER_GET_MAIN_PWR       , "Pwr P %02X %u %u %u %u\n"}
, {c4r35, par4, MEMETER_GET_MAIN_PWR       , "Pwr Q %02X %u %u %u %u\n"}
, {c4r36, par4, MEMETER_GET_MAIN_PWR       , "Pwr S %02X %u %u %u %u\n"}
, {c4ra0, par3, MEMETER_GET_MAIN_V         , "V %02X %u %u %u\n"}
, {c4ra0, par3, MEMETER_GET_MAIN_I         , "I %02X %u %u %u\n"}
, {c4ra0, par4, MEMETER_GET_MAIN_COFF      , "C %02X %u %u %u %u\n"}
, {c4ra0, par1, MEMETER_GET_MAIN_FREQ      , "F %1X %u\n"}
, {c4ra0, par3, MEMETER_GET_MAIN_ANGLE     , "Angle %02X %u %u %u\n"}
, {c4ra0, p442, MEMETER_GET_VARIANT        , "C1A=%s%%"}
, {c4ra0, p443, MEMETER_GET_MEAN_PWR_LAST  , "Addr %05X, %s, mem init %s, not full %s, over %s\n"}
, {c4ra0, par0, MEMETER_GET_TARIF_STATUS   , "tarif stat %02X\n"}
, {c4ra0, par0, MEMETER_GET_LOAD_CTL_STATUS, "load ctl stat %04X\n"}
, {c4ra0, par3, MEMETER_GET_IND_TARIF      , "Ind tarif %02X %02X %02X %02X\n"}
, {c4ra0, par3, MEMETER_GET_IND_PERIOD     , "Ind time %02X %02X %02X %02X\n"}
, {c4ra0, par0, MEMETER_GET_TIMEOUT_COFF   , "Timeout coff %02X\n"}
, {c4ra0, par0, MEMETER_GET_LOSS_ACC_MODE  , "Loss acc mode %04X\n"}
, {c4ra0, par5, MEMETER_GET_LOSS           , "Ploss %u %u %u %u %u %u\n"}
, {c4ra0, par1, MEMETER_GET_MEAN_TIME      , "Mean voltage time %u, mean frequency time %u\n"}
, {c4r52, pac0, MEMETER_ENERGY_RST         , "Energy sum t0"}
, {c4rx1, pac0, MEMETER_ENERGY_RST         , "Energy sum t1"}
, {c4r54, pac0, MEMETER_ENERGY_RST         , "Energy sum t5"}
, {c4rx6, pac0, MEMETER_ENERGY_RST         , "Energy sum t6"}
, {c4rx6, pac0, MEMETER_ENERGY_CUR_YEAR    , "Energy cyr t6"}
, {c4rx6, pac0, MEMETER_ENERGY_LAST_YEAR   , "Energy lyr t6"}
, {c4r61, pac8, MEMETER_ENERGY_MON         , "Energy mon m1 t6"}
, {c4rx6, pac0, MEMETER_ENERGY_CUR_DAY     , "Energy  cd t6"}
, {c4rx6, pac0, MEMETER_ENERGY_LAST_DAY    , "Energy  ld t6"}
, {c4r61, pac8, MEMETER_ENERGY_PHASE       , "Energy ph m1 t6"}
, {c4r62, pac8, MEMETER_ENERGY_PHASE       , "Energy ph m1 t6"}
, {NULL , NULL, 0                          , NULL}
};

static void
usage(int ex) {
    fprintf( stderr
           , "mercumeterq @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "Usage:\n"
             "%s [options] N [params]\n"
             "   -b BAUD         baud rate\n"
             "   -c CNTR         counter addr do query for\n"
             "   -E -X XX        use XX as env name for opt's arg\n"
             "   -f ADDR         src addr in controller\'s dialog\n"
             "   -F FILE         config file\n"
             "   -i TABLE        print INSERT TO TABLE statement in 90 and 91\n"
             "   -I [0123456]    levels for input dumps\n"
             "   -l /dev/line    tty controller is on\n"
             "   -n CNTR         query to CNTR forbiden\n"
             "   -O [0123456]    levels for output dumps\n"
             "   -p CNTR         query to CNTR allowed\n"
             "   -q              silent\n"
             "   -t ADDR         dst addr in controller\'s dialog\n"
             "   -T time         max timeout in 10ms units\n"
             "   -v[v[v...]]     verbose\n"
             "   -w time         tty timeout in 10ms units\n"
             "   -z TEXT         parce TEXT as arguments continuation\n"
             "   -Z TEXT         parce TEXT as arguments continuation with env substitution\n"
             "   -?              this help\n"
             "    N: 0 REQ       request controller with REQ hex string on 2 level\n"
             "       1           query controller addr and query controller on 3 level\n"
             "       2 REQ MBOX  request PLCII modem !CNTR! via controller\'s\n"
             "                     MBOX with REQ hex string on 4 level\n"
             "       3 MBOX      query PLCII modem (CNTR or all) via controller\'s MBOX on 5 level\n"
             "       4 MBOX      query counter (CNTR or all) via controller\'s MBOX on 6 level\n"
           , getprogname()
           )
    ;
    exit(ex);
}

static int
getallc(memeter_cfg *cfg, int report) {
    memeter_answ *w;
    u_int32_t i, k;
    int y, ex = EX_OK;

    do {
        u_int32_t *a;

        if  (!(a = memeter_3exchg(cfg, MEMETER_GET_SEG_NUM))) {
            ifBLIN_QW0("memeter_3exchg");
            break;
        }
        cfg->nmbox = *a;
        free(a);
    } while(0);
    for (i = 0; i < 256; ++i) {
        if  (!(w = memeter_3exchg(cfg, MEMETER_GET_NL_PAGE, i))) {
            ifBLIN_QW0("memeter_3exchg");
            ex = EX_IOERR;
            goto out;
        }
        if  (w->a1 > 32) {
            free(w);
            ifBLIN_QX0("Too long: %u", w->a1);
            ex = EX_PROTOCOL;
            break;
        }
        for (k = 0; k < w->a1; ++k) {
            y = 1;
            if  (cfg->flags & MEMETER_POSIT) y = mercumeter_artst(&cfg->posit, w->a2[k]);
            if  (!!y) y = !mercumeter_artst(&cfg->negat, w->a2[k]);
            if  (!!y && !!mercumeter_aradd(&cfg->meter, w->a2[k])) {
                ifBLIN_QW0("aradd meter %u", w->a2[k]);
                ex = EX_SOFTWARE;
                goto out;
        }   }
        if  (w->a1 != 32) break;
        free(w);
    }
    free(w);
out:
    if  (report) {
        printf("%u devices found:", cfg->meter.idx);
        for (i = 0; i < cfg->meter.idx; ++i) {
            if  (!(i & 3)) printf("\n    ");
            printf(" %08X", cfg->meter.arr[i]);
        }
        printf("\n");
    }
    return(ex);
}

int
main(int argc, char **argv) {
#   define INBUF 512            /* overkill XXXX */
    char *flar = "b:c:Ef:F:i:I:l:n:O:p:qt:T:vw:z:Z:?";
    int c, ex = EX_OK;
    babolo_opts *bos;
    memeter_cfg *cfg;
    const char *o;
    ssize_t len;

    if  (!(cfg = calloc(1, sizeof(memeter_cfg)))) {
        errx(EX_UNAVAILABLE, "Calloc failed #1");
    }
    cfg->flags = BLIN_BIT0 | MEMETER_SCAN;
#   define blin_internal_flags (cfg->flags & BLIN_MASK)
    blin_ctl(BLIN_CTL_FSET, blin_internal_flags);
    blin_ctl(BLIN_CTL_FSET | BLIN_BIT7 | BLIN_CTL_WHAT, BLIN_MODDREL | BLIN_MODDSTR | BLIN_MODTIME);
    if  (!!mercumeter_aradd(&cfg->negat, 0)) {
        ifBLIN_QW0("aradd negat 0");
        ex = EX_SOFTWARE;
        goto out;
    }
    cfg->contrwait = 20000;
    cfg->from = MEMETER_HOST_ADDR;
    cfg->to = MEMETER_BROADDR;
    cfg->baud = B9600;
    cfg->cntr = 0;
    cfg->wait = 10;
    if  (!(bos = babolo_openopts(Bpars_SYME, 'Z'))) {
        ifBLIN_QW0("openopts");
        usage(EX_SOFTWARE);
    }
    if  (babolo_setopts(bos, Bpars_NONU, argc, argv, flar)) {
        ifBLIN_QW0("setopts");
        usage(EX_SOFTWARE);
    }
    while (!!(c = babolo_getopts(bos))) {
        switch (c) {
        case 'b': cfg->baud = babolo_getoptunum(bos, 0);
                ; break;
        case 'c': cfg->cntr = babolo_getoptunum(bos, 16);
                ; cfg->flags &= ~MEMETER_SCAN;
                ; break;
        case 'f': cfg->from = babolo_getoptunum(bos, 16);
                ; break;
        case 'F': {   mife_descriptor *md;
                      babolo_parm *bp;
                      char *txt, *txp;

                  ;   o = babolo_getoptsarg(bos);
                  ;   if  (!(md = mife_init(MIFE_FULL))) {
                          ifBLIN_QW0("mife_init");
                          ex = EX_IOERR;
                          goto out;
                      }
                  ;   if  (0 > mife_ctl(md, MIFE_CTLFILE, o)) {
                          ifBLIN_QW0("MIFE_CTLFILE");
                          ex = EX_IOERR;
                          goto out;
                      }
                  ;   if  (0 > mife_ctl(md, MIFE_CTLREAD, (off_t)0, (ssize_t)0)) {
                          ifBLIN_QW0("MIFE_CTLREAD");
                          ex = EX_IOERR;
                          goto out;
                      }
                  ;   if  (!(txt = mife_point(md, MIFE_POINOFF, (off_t)0))) {
                          ifBLIN_QW0("MIFE_POINOFF");
                          ex = EX_IOERR;
                          goto out;
                      }
                  ;   for (txp = txt; !!*txp; ++txp) if (*txp < ' ') *txp = ' ';
                  ;   if  (!(bp = babolo_getparm( Bpars_NOEN | Bpars_NOAL | Bpars_NONU, &txt, "", 0))) {
                          ifBLIN_QW0("babolo_getparm");
                          ex = EX_IOERR;
                          goto out;
                      }
                  ;   if  (!!babolo_setparm(bos, 0, bp, NULL)) {
                          ifBLIN_QW0("babolo_getparm");
                          ex = EX_DATAERR;
                          goto out;
                  }   }
                ; break;
        case 'i': cfg->tablename = babolo_getoptsarg(bos);
                ; break;
        case 'I': {   size_t i;

                  ;   o = babolo_getoptsarg(bos);
                  ;   for (i = 0; o[i]; ++i) {
                  ;       if  (o[i] >= '0' && o[i] <= '7') {
                  ;           cfg->flags |= MEMETER_0IDUMP >> (o[i] - '0');
                  }   }   }
                ; break;
        case 'l': cfg->linename = babolo_getoptsarg(bos);
                ; break;
        case 'n': if  (!(cfg->flags & MEMETER_NEGAT)) {
                ;     mercumeter_arnul(&cfg->negat);
                ;     cfg->flags |= MEMETER_NEGAT;
                ; }
                ; mercumeter_aradd(&cfg->negat, babolo_getoptunum(bos, 16));
                ; break;
        case 'O': {   size_t i;

                  ;   o = babolo_getoptsarg(bos);
                  ;   for (i = 0; o[i]; ++i) {
                  ;       if  (o[i] >= '0' && o[i] <= '7') {
                  ;           cfg->flags |= MEMETER_0ODUMP >> (o[i] - '0');
                  }   }   }
                ; break;
        case 'p': cfg->flags |= MEMETER_POSIT;
                ; mercumeter_aradd(&cfg->posit, babolo_getoptunum(bos, 16));
                ; break;
        case 'q': BLIN_QUIET(cfg->flags);
                ; break;
        case 't': cfg->to = babolo_getoptunum(bos, 16);
                ; break;
        case 'T': cfg->contrwait = babolo_getoptunum(bos, 0) * 2;
                ; break;
        case 'v': BLIN_VERBOSE(cfg->flags);
                ; break;
        case 'w': cfg->wait = babolo_getoptunum(bos, 0);
                ; break;
        case '?': usage(EX_OK);
        default : usage(EX_USAGE);
    }   }
    errno = 0;
    if  ((o = babolo_getargs(bos))) {
        int type;

        type = strtoul(o, NULL, 0);
        switch (type) {
        case 0:
        ;   errno = 0;
        ;   if  (!(o = babolo_getargs(bos))) goto err;
        ;   {   u_char *cmd, *ans;
                size_t ol;
                int r;

        ;       if  (!(cmd = babolo_hextrin(o, &len))) {
        ;           ifBLIN_QX0("recobe_hextrin");
        ;           ex = EX_SOFTWARE;
        ;           goto out;
        ;       }
        ;       ol = INBUF;
        ;       if  (!(ans = malloc(ol))) {
        ;           ifBLIN_QX0("Malloc failed #1");
        ;           ex = EX_OSERR;
        ;           goto out;
        ;       }
        ;       if  (0 > (r = memeter_2exchg(cfg, cmd, len, ans, &ol))) {
        ;           ifBLIN_QW0("memeter_2exchg");
        ;           goto out;
        ;       } else if (r) {
        ;           printf("answer shortened for %u\n", r);
        ;       }
        ;       memeter_dump(ans, ol, "Got 2 answer ");
        ;   }
        ;   goto out;
        case 1:
        ;   do {
                u_int32_t *a;

        ;       if  (!(a = memeter_3exchg(cfg, MEMETER_GET_ADDR))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       cfg->to = *a;
        ;       printf("found %04X controller\n", *a);
        ;       free(a);
        ;   } while(0);
        ;   do {
                memeter_ansb *v;
                u_int32_t i;

        ;       if  (!(v = memeter_3exchg(cfg, MEMETER_GET_VERINFO))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       printf("version: ");
        ;       if  (v->a1 > 255) {
        ;           ifBLIN_QX0("Too long");
        ;           ex = EX_SOFTWARE;
        ;           break;
        ;       }
        ;       for (i = 0; i < v->a1; ++i) {
        ;           printf("%c", (v->a2[i] < 0x80) ? v->a2[i] : wk[v->a2[i] - 0x80]);
        ;       }
        ;       printf("\n");
        ;       free(v);
        ;   } while(0);
        ;   do {
                memeter_ans7 *d;

        ;       if  (!(d = memeter_3exchg(cfg, MEMETER_GET_TIMEDATE))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       printf( "date: %04u-%02u-%02u %s %02u:%02u:%02u\n"
                      , 2000 + d->a7
                      , 1 + d->a6
                      , 1 + d->a5
                      , wd[d->a4] /* XXXX  a4 */
                      , d->a3
                      , d->a2
                      , d->a1
                      )
        ;       ;
        ;       free(d);
        ;   } while(0);
        ;   do {
                u_int32_t *a, k;

        ;       if  (!(a = memeter_3exchg(cfg, MEMETER_GET_CONFIG))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       for (k = 0; k < 8; ++k, *a >>= 1) {
        ;           printf( "%s %c%s\n"
                          , (!k) ? "config:" : "       "
                          , (*a & 1) ? '+' : '-'
                          , cfb[k]
                    );
        ;       }
        ;       free(a);
        ;   } while(0);
        ;   do {
                u_int32_t *a;

        ;       if  (!(a = memeter_3exchg(cfg, MEMETER_GET_SEG_NUM))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       cfg->nmbox = *a;
        ;       printf("mboxes: %u segments\n", *a);
        ;       free(a);
        ;   } while(0);
        ;   do {
                char *c;
                u_int32_t i;

        ;       if  (!(c = memeter_3exchg(cfg, MEMETER_GET_LOG_FILE))) {
        ;           ifBLIN_QW0("memeter_3exchg");
        ;           break;
        ;       }
        ;       printf("log=");
        ;       for (i = 0; i < 240; ++i) {
        ;           if  ((c[i] >= ' ') && (c[i] < 0x7F)) {
        ;               printf("%c", c[i]);
        ;           } else {
        ;               printf("\\x%02X", c[i]);
        ;       }   }
        ;       printf("\n");
        ;       free(c);
        ;   } while(0);
        ;   ex = getallc(cfg, 1);
        ;   /*  GET_TIMESLOT */
        ;   {   u_int32_t i;

        ;       for (i = 0; i < cfg->meter.idx; ++i) {
        ;           printf("device %u: %08X\n", i, cfg->meter.arr[i]);
        ;           do {
                        memeter_answ *w;
                        u_int32_t *a, k;

        ;               if  (!(w = memeter_3exchg(cfg, MEMETER_GET_P_LIST, cfg->meter.arr[i]))) {
        ;                   ifBLIN_QW0("memeter_3exchg");
        ;                   break;
        ;               }
        ;               if  (!(a = memeter_3exchg(cfg, MEMETER_GET_PARENT, cfg->meter.arr[i]))) {
        ;                   ifBLIN_QW0("memeter_3exchg");
        ;                   break;
        ;               }
        ;               if  (w->a1 > 16) {
        ;                   ifBLIN_QX0("Too long");
        ;                   ex = EX_SOFTWARE;
        ;                   break;
        ;               }
        ;               for (k = 0; k < w->a1; ++k) {
        ;                    if  (!(k & 3)) printf(k ? "\n             " : "parent nodes ");
        ;                    printf( " %c%08X%c"
                                   , (w->a2[k] == *a) ? '+' : ' '
                                   , w->a2[k]
                                   , (w->a2[k] == *a) ? '+' : ' '
                                   )
        ;                    ;
        ;               }
        ;               printf("\n");
        ;               free(w);
        ;               free(a);
        ;           } while(0);
        ;           do {
                        memeter_ans3 *g;
                        u_int32_t *a, k;

        ;               if  (!(g = memeter_3exchg(cfg, MEMETER_GET_MOD_CONFIG, cfg->meter.arr[i]))) {
        ;                   ifBLIN_QW0("memeter_3exchg");
        ;                   break;
        ;               }
        ;               if  (!(a = memeter_3exchg(cfg, MEMETER_GET_SL_CONFIG, cfg->meter.arr[i]))) {
        ;                   ifBLIN_QW0("memeter_3exchg");
        ;                   break;
        ;               }
        ;               printf("PLC Ver: %3u, Host proto: %3u, BCNF:", g->a1, g->a2);
        ;               for ( k = 0
                            ; (k < 8) && ((k < 2) || (g->a3) || *a)
                            ; ++k, g->a3 >>= 1, *a >>= 1
                            ) {
        ;                   printf( "%s %c%s %s %c%s\n"
                                  , (!k) ? " " : "                                     "
                                  , (g->a3 & 1) ? '+' : '-'
                                  , csm[k]
                                  , (!k) ? "slave config:" : "             "
                                  , (*a & 1) ? '+' : '-'
                                  , ccn[k]
                            );
        ;               }
        ;               free(g);
        ;               free(a);
        ;           } while(0);
        ;           {   u_int32_t m;

        ;               for (m = 0; m < cfg->nmbox; ++m) {
        ;                   do {
                                memeter_ans3 *g;
                                u_int32_t k;

        ;                       g = memeter_3exchg(cfg, MEMETER_GET_SEG_STATUS, cfg->meter.arr[i], m);
        ;                       if  (!g) {
        ;                           ifBLIN_QW0("memeter_3exchg");
        ;                           break;
        ;                       }
        ;                       printf( "  mbox %3u, cicle %3u0ms, %5u errors, status %s"
                                      , m, g->a2, g->a3, cbx[g->a1 & MEMETER_SEG_STATUS_MASK]
                                      )
        ;                       ;
        ;                       for ( k = 16
                                    ; (k < 20) && ((k < 18) || (g->a1 & 0xF0))
                                    ; ++k, g->a1 >>= 1
                                    ) {
        ;                           printf( "\n%s %c%s"
                                          , (!k) ? "" : "                                             "
                                          , (g->a1 & 0x10) ? '+' : '-'
                                          , cbx[k]
                                    );
        ;                       }
        ;                       printf("\n");
        ;                   } while(0);
        ;                   do {
                                memeter_ansb *b;
                                u_int32_t k;

        ;                       b = memeter_3exchg(cfg, MEMETER_GET_SEG_REQ, cfg->meter.arr[i], m);
        ;                       if  (!b) {
        ;                           ifBLIN_QW0("memeter_3exchg");
        ;                           break;
        ;                       }
        ;                       printf( "    req len=%04X(%u)", b->a1, b->a1);
        ;                       for (k = 0; k < b->a1; ++k) {
        ;                           if  (!(k & 0x0F)) printf("\n      %04X", k);
        ;                           if  (!(k & 0x01)) printf(" ");
        ;                           if  (!(k & 0x07)) printf(" ");
        ;                           printf("%02X", b->a2[k]);
        ;                       }
        ;                       printf("\n");
        ;                   } while(0);
        ;                   do {
                                memeter_ans8 *b;
                                u_int32_t k;

        ;                       b = memeter_3exchg(cfg, MEMETER_GET_SEG_ANS, cfg->meter.arr[i], m);
        ;                       if  (!b) {
        ;                           ifBLIN_QW0("memeter_3exchg");
        ;                           break;
        ;                       }
        ;                       printf( "    ans len=%04X(%u) date: %04u-%02u-%02u %02u:%02u:%02u"
                                      , b->a7
                                      , b->a7
                                      , 2000 + b->a6
                                      , 1 + b->a5
                                      , 1 + b->a4
                                      , b->a3
                                      , b->a2
                                      , b->a1
                                      )
        ;                       ;
        ;                       for (k = 0; k < b->a7; ++k) {
        ;                           if  (!(k & 0x0F)) printf("\n      %04X", k);
        ;                           if  (!(k & 0x01)) printf(" ");
        ;                           if  (!(k & 0x07)) printf(" ");
        ;                           printf("%02X", b->a8[k]);
        ;                       }
        ;                       printf("\n");
        ;                   } while(0);
        ;   }   }   }   }
        ;   goto out;
        case 2:
        ;   errno = 0;
        ;   if  (cfg->flags & MEMETER_SCAN) {
        ;       ifBLIN_QX0("No -c CNTR");
        ;       goto err;
        ;   }
        ;   if  (!(o = babolo_getargs(bos))) {
        ;       ifBLIN_QX0("No hex string");
        ;       goto err;
        ;   }
        ;   {   u_char *cmd, *ans;
                memeter_ctr ctr;
                size_t ol;
                int r, q;

        ;       if  (!(cmd = babolo_hextrin(o, &len))) {
        ;           ifBLIN_QX0("recobe_hextrin");
        ;           ex = EX_SOFTWARE;
        ;           goto out;
        ;       }
        ;       if  (cfg->flags & MEMETER_SCAN) {
        ;           ifBLIN_QX0("Counter number absent");
        ;           ex = EX_OSERR;
        ;           goto out;
        ;       }
        ;       ctr.ctr = cfg->cntr;
        ;       if  (!(o = babolo_getargs(bos))) {
        ;           ifBLIN_QX0("No mbox N");
        ;           goto err;
        ;       }
        ;       ctr.mbx = strtoul(o, NULL, 0);
        ;       ol = INBUF;
        ;       if  (!(ans = malloc(ol))) {
        ;           ifBLIN_QX0("Malloc failed #1");
        ;           ex = EX_OSERR;
        ;           goto out;
        ;       }
        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD | MEMETER_SEG_STATUS_RETYPE;
        ;       if  (0 > (q = memeter_4req(&ctr, cmd, len))) {
        ;           ifBLIN_QW0("memeter_4req");
        ;           goto out;
        ;       }
        ;       do {
        ;           ++q;
        ;           usleep(q * MEMETER_SLEEPTICK);
        ;           if  (0 > (r = memeter_4ans(&ctr, ans, &ol))) {
        ;               ifBLIN_QX0("memeter_4ans %d", r);
        ;               ex = EX_OSERR;
        ;               goto out;
        ;           } else if (!r) {
        ;               memeter_dump(ans, ol, "Got 4 answer ");
        ;               goto out;
        ;           }
        ;           q = q * 2 + r;
        ;       } while (q < cfg->contrwait);
        ;       ifBLIN_QX0("memeter_4ans %d %d", r, q);
        ;   }
        ;   goto out;
        case 3:
        ;   errno = 0;
        ;   if  (!(o = babolo_getargs(bos))) {
        ;       ifBLIN_QX0("No mbox num");
        ;       goto err;
        ;   }
        ;   {
                u_int32_t i;
                memeter_ctr ctr;
                int q, r, qq;
                u_int32_t k;
                u_char *ans;

        ;       ctr.mbx = strtoul(o, NULL, 0);
        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD | MEMETER_SEG_STATUS_RETYPE;
        ;       if  (cfg->flags & MEMETER_SCAN) {
        ;           ex = getallc(cfg, 0);
        ;           if  (!cfg->meter.idx) {
        ;               ifBLIN_QX0("No counters found");
        ;               ex = EX_IOERR;
        ;               break;
        ;       }   }
        ;       for (i = 0;; ++i) {
        ;           if  (cfg->flags & MEMETER_SCAN) {
        ;               if  (i >= cfg->meter.idx) break;
        ;               ctr.ctr = cfg->meter.arr[i];
        ;           } else {
        ;               if (!!i) break;
        ;               ctr.ctr = cfg->cntr;
        ;           }
        ;           qq = MEMETER_MINWAIT;
        ;           if  (0 > (q = memeter_5req(&ctr, MEMETER_PLCII_GETCNF))) {
        ;               ifBLIN_QX0("memeter_5req %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           if  ((r = memeter_5ans(&ctr, MEMETER_PLCII_GETCNF, &ans, qq))) {
        ;               ifBLIN_QX0("memeter_5ans %08X(%d)", r, r);
        ;               ex = EX_IOERR;
        ;           } else {
        ;               if  (cfg->meter.idx) printf("(((( %08X ))))\n", ctr.ctr);
        ;               printf("Got modem config %02X %02X", ans[0], ans[1]);
        ;               for (k = 0; (k < 8) && ((k < 1) || ans[2]); ++k, ans[2] >>= 1) {
        ;                   printf( "%s %c%s\n"
                                  , (!k) ? "" : "                      "
                                  , (ans[2] & 1) ? '+' : '-'
                                  , csm[k]
                            );
        ;           }   }
        ;           qq = MEMETER_MINWAIT;
        ;           if  (0 > (q = memeter_5req(&ctr, MEMETER_PLCII_GETCNT))) {
        ;               ifBLIN_QX0("memeter_5req %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           if  ((r = memeter_5ans(&ctr, MEMETER_PLCII_GETCNT, &ans, qq))) {
        ;               ifBLIN_QX0("memeter_5ans %08X(%d)", r, r);
        ;               ex = EX_IOERR;
        ;           } else {
        ;               if  (cfg->meter.idx) printf("(((( %08X ))))\n", ctr.ctr);
        ;               printf( "Got node counters MCNT=%02X HCNT=%02X LCNT=%02X\n"
                              , ans[0], ans[1], ans[2]
                              )
        ;               ;
        ;   }   }   }
        ;   goto out;
        case 4:
        ;   errno = 0;
        ;   if  (!(o = babolo_getargs(bos))) {
        ;       ifBLIN_QX0("No mbox num");
        ;       goto err;
        ;   }
        ;   {
                u_int32_t i, j;
                memeter_ctr ctr;
                u_int32_t *ans;
                int q, r, qq;

        ;       ctr.mbx = strtoul(o, NULL, 0);
        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD | MEMETER_SEG_STATUS_RETYPE;
        ;       if  (cfg->flags & MEMETER_SCAN) {
        ;           ex = getallc(cfg, 0);
        ;           if  (!cfg->meter.idx) {
        ;               ifBLIN_QX0("No counters found");
        ;               ex = EX_IOERR;
        ;       }   }
        ;       for (i = 0;; ++i) {
        ;           if  (cfg->flags & MEMETER_SCAN) {
        ;               if  (i >= cfg->meter.idx) break;
        ;               ctr.ctr = cfg->meter.arr[i];
        ;           } else {
        ;               if (!!i) break;
        ;               ctr.ctr = cfg->cntr;
        ;           }
        ;           for (j = 0; !!cmd4[j].fmt; ++j) {
        ;               qq = MEMETER_MINWAIT;
        ;               if  (0 > (q = cmd4[j].req(&ctr, cmd4[j].cmd))) {
        ;                  ifBLIN_QX0("memeter_6req %d", q);
        ;                  ex = EX_IOERR;
        ;               }
        ;               if  (qq < q) qq = q;
        ;               if  ((r = memeter_6ans(&ctr, cmd4[j].cmd, &ans, qq))) {
        ;                   ifBLIN_QX0("memeter_6ans[%u,%u] %08X(%d)", i, j, r, r);
        ;                   ex = EX_IOERR;
        ;               } else {
        ;                   if  (cfg->flags & MEMETER_SCAN) printf("(((( %08X ))))\n", ctr.ctr);
        ;                   cmd4[j].pr(cmd4[j].fmt, ans);
        ;   }   }   }   }
        ;   goto out;
        case 90:
        ;   {   memeter_ctr ctr;
                int q, qq;
                u_int32_t i;

        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_REMASK | MEMETER_SEG_STATUS_TRANS_TYPE;
        ;       ex = getallc(cfg, 0);
        ;       if  (!cfg->meter.idx) {
        ;           ifBLIN_QX0("No counters found");
        ;           ex = EX_IOERR;
        ;           break;
        ;       }
        ;       for (qq = MEMETER_MINWAIT, i = 0; i < cfg->meter.idx; ++i) {
        ;           ctr.ctr = cfg->meter.arr[i];
        ;           ctr.mbx = 0;
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_GET_TIME, 0))) {
        ;               ifBLIN_QX0("memeter_6req GET_TIME %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 1;
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_GET_TRANS_COFF))) {
        ;               ifBLIN_QX0("memeter_6req GET_TRANS_COFF %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 2;
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_ENERGY_RST, 0))) {
        ;               ifBLIN_QX0("memeter_6req ENERGY_RST %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 3;
        ;           q = memeter_6req(&ctr, MEMETER_GET_MAIN_1PWR, 0, MEMETER_C_PARAM_PWR_P);
        ;           if  (0 > q) {
        ;               ifBLIN_QX0("memeter_6req GET_MAIN_1PWR %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;       }
        ;       ctr.mst = MEMETER_SEG_STATUS_RETYPE | MEMETER_SEG_STATUS_RECMD;
        ;       {   u_int32_t a;

        ;           for (a = 4; a < cfg->nmbox; ++a) {
        ;               for (ctr.mbx = a, i = 0; i < cfg->meter.idx; ++i) {
        ;                   ctr.ctr = cfg->meter.arr[i];
        ;                   if  (0 > (q = memeter_4req(&ctr, NULL, 0))) {
        ;                       ifBLIN_QX0("memeter_4req %d", q);
        ;                       ex = EX_IOERR;
        ;       }   }   }   }
        ;       for (i = 0; i < cfg->meter.idx; ++i) {
                    u_int32_t *a0, *a1, *a2, *a3;
                    int r;

        ;           ctr.ctr = cfg->meter.arr[i];
        ;           ctr.mbx = 0;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_TIME, &a0, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_TIME %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 1;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_TRANS_COFF, &a1, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_TRANS_COFF %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 2;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_ENERGY_RST, &a2, qq))) {
        ;               ifBLIN_QX0("memeter_6ans ENERGY_RST %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 3;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_MAIN_1PWR, &a3, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_MAIN_1PWR %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           if  (cfg->tablename) {
        ;               printf( "INSERT INTO %s"
                                " VALUES(\'20%02x-%02x-%02x %02x:%02x:%02x\'"
                                ", \'%08X\', %u.%03u, %u.%02u);\n"
                              , cfg->tablename
                              , a0[6], a0[5], a0[4], a0[2], a0[1], a0[0]
                              , ctr.ctr
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] / 1000)
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] % 1000)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] / 100)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] % 100)
                              )
        ;               ;
        ;           } else {
        ;               printf( "20%02x-%02x-%02x_%02x:%02x:%02x\t%08X\t%u.%03u\t%u.%02u\n"
                              , a0[6], a0[5], a0[4], a0[2], a0[1], a0[0]
                              , ctr.ctr
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] / 1000)
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] % 1000)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] / 100)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] % 100)
                              )
        ;               ;
        ;   }   }   }
        ;   goto out;
        case 91:
        ;   {   memeter_ctr ctr;
                int q, qq = 0;
                u_int32_t i;

        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD
                        | MEMETER_SEG_STATUS_RETYPE
                        | MEMETER_SEG_STATUS_TRANS_TYPE
        ;       ;
        ;       ex = getallc(cfg, 0);
        ;       if  (!cfg->meter.idx) {
        ;           ifBLIN_QX0("No counters found");
        ;           ex = EX_IOERR;
        ;           break;
        ;       }
        ;       for (qq = MEMETER_MINWAIT, i = 0; i < cfg->meter.idx; ++i) {
        ;           ctr.ctr = cfg->meter.arr[i];
        ;           ctr.mbx = 0;
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_GET_TIME, 0))) {
        ;               ifBLIN_QX0("memeter_6req GET_TIME %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 1;
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_GET_TRANS_COFF))) {
        ;               ifBLIN_QX0("memeter_6req GET_TRANS_COFF %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 2;
        ;           ctr.ctr = cfg->meter.arr[i];
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_ENERGY_RST, 0))) {
        ;               ifBLIN_QX0("memeter_6req ENERGY_RST %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;           ctr.mbx = 3;
        ;           ctr.ctr = cfg->meter.arr[i];
        ;           q = memeter_6req(&ctr, MEMETER_GET_MAIN_1PWR, 0, MEMETER_C_PARAM_PWR_P);
        ;           if  (0 > q) {
        ;               ifBLIN_QX0("memeter_6req GET_MAIN_1PWR %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;       }
        ;       for (i = 0; i < cfg->meter.idx; ++i) {
                    u_int32_t *a0, *a1, *a2, *a3;
                    int r;

        ;           ctr.ctr = cfg->meter.arr[i];
        ;           ctr.mbx = 0;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_TIME, &a0, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_TIME %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 1;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_TRANS_COFF, &a1, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_TRANS_COFF %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 2;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_ENERGY_RST, &a2, qq))) {
        ;               ifBLIN_QX0("memeter_6ans ENERGY_RST %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           ctr.mbx = 3;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_MAIN_1PWR, &a3, qq))) {
        ;               ifBLIN_QX0("memeter_6ans GET_MAIN_1PWR %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           if  (cfg->tablename) {
        ;               printf( "INSERT INTO %s"
                                " VALUES(\'20%02x-%02x-%02x %02x:%02x:%02x\'"
                                ", \'%08X\', %u.%03u, %u.%02u);\n"
                              , cfg->tablename
                              , a0[6], a0[5], a0[4], a0[2], a0[1], a0[0]
                              , ctr.ctr
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] / 1000)
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] % 1000)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] / 100)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] % 100)
                              )
        ;               ;
        ;           } else {
        ;               printf( "20%02x-%02x-%02x_%02x:%02x:%02x\t%08X\t%u.%03u\t%u.%02u\n"
                              , a0[6], a0[5], a0[4], a0[2], a0[1], a0[0]
                              , ctr.ctr
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] / 1000)
                              , (u_int32_t)((u_int64_t)a2[1] * a1[0] * a1[1] % 1000)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] / 100)
                              , (u_int32_t)((u_int64_t)a3[1] * a1[0] * a1[1] % 100)
                              )
        ;               ;
        ;   }   }   }
        ;   goto out;
        case 92:
        ;   errno = 0;
        ;   if  (!(o = babolo_getargs(bos))) {
        ;       ifBLIN_QX0("No month num");
        ;       goto err;
        ;   }
        ;   {   memeter_ctr ctr;
                u_int32_t i, m;
                int q, qq = 0;

        ;       m = strtoul(o, NULL, 0);
        ;       if  ((m < 1) || (m > 12)) {
        ;           ifBLIN_QX0("Invalid month num %u", m);
        ;           goto err;
        ;       }
        ;       ctr.cfg = cfg;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD | MEMETER_SEG_STATUS_RETYPE;
        ;       ex = getallc(cfg, 0);
        ;       if  (!cfg->meter.idx) {
        ;           ifBLIN_QX0("No counters found");
        ;           ex = EX_IOERR;
        ;           break;
        ;       }
        ;       for (qq = MEMETER_MINWAIT, ctr.mbx = 11, i = 0; i < cfg->meter.idx; ++i) {
        ;           ctr.ctr = cfg->meter.arr[i];
        ;           if  (0 > (q = memeter_6req(&ctr, MEMETER_ENERGY_MON, 0, m))) {
        ;               ifBLIN_QX0("memeter_6req %d", q);
        ;               ex = EX_IOERR;
        ;           }
        ;           if  (qq < q) qq = q;
        ;       }
        ;       for (i = 0; i < cfg->meter.idx; ++i) {
                    u_int32_t *a1;
                    int r;

        ;           ctr.ctr = cfg->meter.arr[i];
        ;           ctr.mbx = 11;
        ;           if  ((r = memeter_6ans(&ctr, MEMETER_ENERGY_RST, &a1, qq))) {
        ;               ifBLIN_QX0("memeter_6ans %08X(%d)", r, r);
        ;               continue;
        ;           }
        ;           printf("%s %u.%03u\n", mm[m - 1], a1[1] / 1000, a1[1] % 1000);
        ;   }   }
        ;   goto out;
        case 99:
        ;   errno = 0;
        ;   if  (!(o = babolo_getargs(bos))) {
        ;       ifBLIN_QX0("No Ki");
        ;       goto err;
        ;   }
        ;   {   memeter_ctr ctr;
                u_int32_t m;
                int q;

        ;       m = strtoul(o, NULL, 0);
        ;       ctr.mbx = 11;
        ;       ctr.cfg = cfg;
        ;       ctr.ctr = cfg->cntr;
        ;       ctr.mst = MEMETER_SEG_STATUS_RECMD | MEMETER_SEG_STATUS_RETYPE;
        ;       ex = getallc(cfg, 0);
        ;       if  (!cfg->meter.idx) {
        ;           ifBLIN_QX0("No counters found");
        ;           ex = EX_IOERR;
        ;           break;
        ;       }
        ;       if  (0 > (q = memeter_6req(&ctr, MEMETER_SET_TRANS_COFF, 1, m))) {
        ;           ifBLIN_QX0("memeter_6req %d", q);
        ;           ex = EX_IOERR;
        ;       }
        ;       if  (q < MEMETER_MINWAIT) q = MEMETER_MINWAIT;
        ;       {   u_int32_t *a1;
                    int r;

        ;           if  ((r = memeter_6ans(&ctr, MEMETER_SET_TRANS_COFF, &a1, q))) {
        ;               ifBLIN_QX0("memeter_6ans %08X(%d)", r, r);
        ;       }   }
        ;       if  (0 > (q = memeter_6req(&ctr, MEMETER_GET_TRANS_COFF))) {
        ;           ifBLIN_QX0("memeter_6req %d", q);
        ;           ex = EX_IOERR;
        ;       }
        ;       if  (q < MEMETER_MINWAIT) q = MEMETER_MINWAIT;
        ;       {   u_int32_t *ans;
                    int r;

        ;           if  ((r = memeter_6ans(&ctr, MEMETER_GET_TRANS_COFF, &ans, q))) {
        ;               ifBLIN_QX0("memeter_6ans %08X(%d)", r, r);
        ;           }
        ;           printf("Ku=%u Ki=%u\n", ans[0], ans[1]);
        ;   }   }
        ;   goto out;
    }   }
err:
    if  (errno) {
        ifBLIN_QW0("babolo_getarg #1");
        ex = EX_SOFTWARE;
        goto out;
    } else {
        ifBLIN_QX0("Nothing to do");
        ex = EX_USAGE;
        goto out;
    }
out:
    exit(ex);
}
